/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 565 $ $LastChangedDate: 2002-08-25 20:18:26 +0200 (Sun, 25 Aug 2002) $
 *
 * Ast - Utility function for manipulating Gont Abstract Synatx Trees.
 */
 
open Ast;
open Util;

string simple_name(<<'a>symbol>bindable s)
{
    switch s {
    case SimpleName[x]:
	return x;
	
    case QualifiedName[_, x]:
	return x;
	
    case Symbol[x]:
	return x.name;
    }
}

string symbol_name(<'a>symbol x)
{
    if (x.module != "")
	return x.module + "::" + x.name;
    else
	return x.name;
}

string qualified_name(<<'a>symbol>bindable s)
{
    switch s {
    case SimpleName[x]:
	return x;
	
    case QualifiedName[m, n]:
	return m + "::" + n;
	
    case Symbol[x]:
	return symbol_name(x);
    }
}

<<'a>symbol>bindable rev_qualified_name(string name)
{
    <string>List::t l = String::split("::", name);
    
    switch l {
    case List::Cons[m, List::Cons[n, List::Nil]]:
	return QualifiedName[m, n];
    case List::Cons[n, List::Nil]:
	return SimpleName[n];
    case _:
	raise Invalid_argument["Ast::rev_qualified_name"];
    }
}

void set_property(var_symbol s, var_symbol_property p)
{
    s.a.properties = List::Cons[p, s.a.properties];
}

bool has_property(var_symbol s, var_symbol_property p)
{
    return List::exists((fun bool (var_symbol_property x) 
		  		(Hacks::union_tag(x) == Hacks::union_tag(p))),
			s.a.properties);
}

ty type_map(*(ty (ty)) f, ty t)
{
    ty self(ty tt)
    {
	return type_map(f, tt);
    }

    <'a>field self_fld(<'a>field fld)
    {
	ty tt;

	push_location(fld.fld_loc);
	try {
	    tt = self(fld.fld_type);
	} finally {
	    pop_location();
	}
	
	return ({fld_type = tt,
	     fld_symbol = fld.fld_symbol,
	     fld_loc = fld.fld_loc});
    }
    
    t = f(t);
    
    switch t {
    case Tuple[lst]:
	return Tuple[List::map(self, lst)];
    
    case Func[ft]:
	*[ty, <id>option] map_fnc_arg(*[ty, <id>option] p)
	{
	    let [tt, nn] = p in
	    return [self(tt), nn];
	}
	
	return Func[({ret_type = self(ft.ret_type),
		      parms = List::map(map_fnc_arg, ft.parms),
		      tyvars = ft.tyvars})];
    
    case Struct[flds]:
	return Struct[List::map(self_fld, flds)];
	
    case OptStruct[flds]:
	return OptStruct[List::map(self_fld, flds)];
	
    case Union[flds]:
	return Union[List::map(self_fld, flds)];
    
    case NamedType[nt]:
	return NamedType[({parms = List::map(self, nt.parms),
		   	   tydef = nt.tydef})];

    case XTyVar[_]:
	return t;   // ???

    case _:
	return t;
    }
}

ty copy_type(ty t)
{
    ty f(ty tt)
    {
	switch tt {
	case XTyVar[x]:
	    switch x.t {
	    case Some[y]:
		return XTyVar[({t = Some[copy_type(y)], name = x.name})];
	    case None:
		return XTyVar[({t = None, name = x.name})];
	    }
	case _:
	    return tt;
	}
    }

    return type_map(f, t);
}

void type_iter(*(void (ty)) f, ty t)
{
    void self(ty tt)
    {
	type_iter(f, tt);
    }

    void self_fld(<'a>field fld)
    {
	push_location(fld.fld_loc);
	try {
	    self(fld.fld_type);
	} finally {
	    pop_location();
	}
    }
    
    f(t);
    
    switch t {
    case Tuple[lst]:
	List::iter(self, lst);
    
    case Func[ft]:
	void map_fnc_arg(*[ty, <id>option] p)
	{
	    let [tt, _] = p in
	    self(tt);
	}
	
	self(ft.ret_type);
	List::iter(map_fnc_arg, ft.parms);
    
    case Struct[flds]:
	List::iter(self_fld, flds);
	
    case OptStruct[flds]:
	List::iter(self_fld, flds);
	
    case Union[flds]:
	List::iter(self_fld, flds);
    
    case NamedType[nt]:
	List::iter(self, nt.parms);
    
    case _:
	skip;
    }
}

// vim:sts=4:sw=4

