/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 571 $ $LastChangedDate: 2002-08-25 22:24:49 +0200 (Sun, 25 Aug 2002) $
 *
 * Xtcheck - check whether there are none weak type variables.
 */

open Ast;
open Util;
open Type;

void xtcheck_type(ty t)
{
    if (has_weak_tyvars(t))
	error("weak type variables not allowed here (type is " + 
	      string_of_type(t) + ")");
}

void xtcheck_raw_expr(nl_expr ex)
{
    switch ex.raw {
    case SimpleBinOp[_, e1, e2]:
	xtcheck_expr(e1);
	xtcheck_expr(e2);
	
    case SimpleUnOp[_, e]:
	xtcheck_expr(e);
    
    case Ref[_]:
	skip;

    case Call[e, el]:
	xtcheck_expr(e);
	List::iter(xtcheck_expr, el);

    case TupleExpr[el]:
	List::iter(xtcheck_expr, el);

    case UnionExpr[u]:
	xtcheck_expr(u.nx_expr);
    
    case Lambda[l]:
	xtcheck_lambda(l);

    case StructExpr[ss]:
	List::iter(fun void (sini_expr s) (xtcheck_expr(s.sx_expr)), ss);

    case DotExpr[d]:
	xtcheck_expr(d.sx_expr);
    
    case BoolConst[_]:
    case IntConst[_]:
    case FloatConst[_]:
    case StringConst[_]:
    case VoidConst:
    case NullConst:
	skip;
    }
}

void xtcheck_expr(expr e)
{
    locate(xtcheck_raw_expr, e);
}

void xtcheck_nl_pattern(nl_pattern p)
{
    switch p {
    case VoidPattern:
	skip;
	
    case VarPattern[v]:
	xtcheck_type(v.var_type);
    
    case ExprPattern[e]:
	xtcheck_expr(e);

    case TuplePattern[lst]:
	List::iter(xtcheck_pattern, lst);
    
    case UnionPattern[np]:
	xtcheck_pattern(np.np_pattern);
    }
}

void xtcheck_pattern(pattern p)
{
    locate(xtcheck_nl_pattern, p);
}

void xtcheck_nl_stmt(nl_stmt s)
{
    void xtcheck_case_clause(case_clause cc)
    {
	List::iter(xtcheck_pattern, List::map(fst, cc.cases));
	List::iter(xtcheck_expr, List::map(snd, cc.cases));
	xtcheck_stmt(cc.body);
    }
    
    switch s {
    case Skip:
	skip;
    
    case Expr[e]:
	xtcheck_expr(e);
    
    case Return[e]:
	xtcheck_expr(e);

    case VarDef[v]:
	xtcheck_type(v.var_type);
	
    case Block[lst]:
	List::iter(xtcheck_stmt, lst);

    case Pair[s1, s2]:
	xtcheck_stmt(s1);
	xtcheck_stmt(s2);
    
    case Switch[e, clauses]:
	xtcheck_expr(e);
	List::iter(xtcheck_case_clause, clauses);
	
    case Let[assigns, st]:
	void check_one(*[pattern, expr] a)
	{
	    let [p, e] = a in {
		xtcheck_expr(e);
		xtcheck_pattern(p);
	    }
	}

	List::iter(check_one, assigns);
	xtcheck_stmt(st);
    
    case If[e, s1, s2]:
	xtcheck_expr(e);
	xtcheck_stmt(s1);
	xtcheck_stmt(s2);
    
    case Loop[l]:
	xtcheck_stmt(l.ll_body);
	xtcheck_stmt(l.ll_inc);
    
    case Break[_]:
	skip;
	
    case BreakIfNot[e]:
	xtcheck_expr(e);
    
    case Open[_]:
	skip;
    
    case Try[tr]:
	xtcheck_stmt(tr.try_body);
	List::iter(xtcheck_case_clause, tr.try_with);
	xtcheck_stmt(tr.try_finally);
    
    case Raise[e]:
	xtcheck_expr(e);
    }
}

void xtcheck_stmt(stmt s)
{
    locate(xtcheck_nl_stmt, s);
}

void xtcheck_lambda(lambda_record l)
{
    List::iter(fun void (var_symbol s) (xtcheck_type(s.stype)), 
	       l.lambda_parms);
    xtcheck_type(l.lambda_type);
    xtcheck_stmt(l.lambda_body);
}

void xtcheck_nl_decl(nl_decl nld)
{
    switch nld {
    case FuncDecl[fn]:
	xtcheck_lambda(fn.fn_lambda);
    
    case SectionDecl[_, l]:
	xtcheck_lambda(l);
    
    case GlobalDecl[v]:
    case ExternC[v]:
    case FuncProto[v]:
    case XCFuncProto[_, v]:
    case ExceptionDecl[v]:
	xtcheck_type(v.var_type);
	
    case TyDef[ts]:
	xtcheck_type(ts.stype);
	
    case TopLevelOpen[_]:
    case DummyDecl:
	skip;
    }
}

void xtcheck_decl(decl d)
{
    locate(xtcheck_nl_decl, d);
}

// vim:sts=4:sw=4
